/////////////////////////////////////////////////////////////////////////
//	"ext2lib.h"
//------------------------------------------------------------------------
//	$Id: ext2lib.h 56 2003-12-11 14:12:07Z  $
//////////////////////////////////////////////////////////////////////////
#ifndef	EXT2LIB_H_INCLUDED
#define	EXT2LIB_H_INCLUDED

#define	EXT2LIB_CFG_LE

#if	defined(WIN32)
# include	"win32/devctl.h"
# include	<tchar.h>
#endif
#if	defined(LINUX)
# include	"linux/devctl.h"
#endif

#if	!defined(ATTB_PACKED)
# if	defined(LINUX)
#   define	ATTB_PACKED	__attribute ((packed))
# endif
# if	defined(WIN32)
#   define	ATTB_PACKED
# endif
#endif

class	Ext2Lib
{
public:
	enum	ext2lib_enum
	{
		BLOCK_SIZE           = 1024,
		CHECK_SUPERBLOCK_NUM = 8,
		EXT2_SUPER_MAGIC     = 0xEF53,
	};
	enum os_enum
	{
		OS_LINUX    = 0,
		OS_HURD     = 1,
		OS_MASIX    = 2,
		OS_FREEBSD  = 3,
		OS_LITES    = 4,
	};

#if	defined(_MSC_VER)
#	pragma pack(push,1)
#endif
	typedef unsigned char		Uint8;
	typedef signed char			Sint8;
	typedef unsigned short		Uint16;
	typedef signed short		Sint16;
	typedef unsigned int		Uint32;
	typedef int					Sint32;
#if	defined(LINUX)
	typedef unsigned long long	Uint64;
	typedef long long			Sint64;
#else
	typedef unsigned __int64	Uint64;
	typedef __int64				Sint64;
#endif
	typedef Uint8				dstring;

	struct ext2_super_block
	{
		Uint32		s_inodes_count;			/* Inodes count */
		Uint32		s_blocks_count;			/* Blocks count */
		Uint32		s_r_blocks_count;		/* Reserved blocks count */
		Uint32		s_free_blocks_count;	/* Free blocks count */
		Uint32		s_free_inodes_count;	/* Free inodes count */
		Uint32		s_first_data_block;		/* First Data Block */
		Uint32		s_log_block_size;		/* Block size */
		Sint32		s_log_frag_size;		/* Fragment size */
		Uint32		s_blocks_per_group;		/* # Blocks per group */
		Uint32		s_frags_per_group;		/* # Fragments per group */
		Uint32		s_inodes_per_group;		/* # Inodes per group */
		Uint32		s_mtime;				/* Mount time */
		Uint32		s_wtime;				/* Write time */
		Uint16		s_mnt_count;			/* Mount count */
		Sint16		s_max_mnt_count;		/* Maximal mount count */
		Uint16		s_magic;				/* Magic signature */
		Uint16		s_state;				/* File system state */
		Uint16		s_errors;				/* Behaviour when detecting errors */
		Uint16		s_minor_rev_level;		/* minor revision level */
		Uint32		s_lastcheck;			/* time of last check */
		Uint32		s_checkinterval;		/* max. time between checks */
		Uint32		s_creator_os;			/* OS */
		Uint32		s_rev_level;			/* Revision level */
		Uint16		s_def_resuid;			/* Default uid for reserved blocks */
		Uint16		s_def_resgid;			/* Default gid for reserved blocks */
	} ATTB_PACKED;
#if	defined(_MSC_VER)
#	pragma pack(pop)
#endif
	static Uint32	e32(Uint32 toswap) { return (toswap << 24) | ((toswap << 8) & 0x00ff0000) | ((toswap >> 8) & 0x0000ff00) | (toswap >> 24); }
	static Uint16	e16(Uint16 toswap) { return (toswap << 8) | (toswap >> 8); }
#if defined(EXT2LIB_CFG_LE)
	// For little endian machines (x86)
	static Sint32	se32(Uint32 x) { return Ext2Lib::e32(x); }
	static Sint16	se16(Uint16 x) { return Ext2Lib::e16(x); }
	static Uint32	ue32(Uint32 x) { return x; }
	static Uint16	ue16(Uint16 x) { return x; }
#else
	// For big endian machines
	static Sint32	se32(Uint32 x) { return x; }
	static Sint16	se16(Uint16 x) { return x; }
	static Uint32	ue32(Uint32 x) { return Ext2Lib::endian32(x); }
	static Uint16	ue16(Uint16 x) { return Ext2Lib::endian16(x); }
#endif

private:
	DevCtl						*devCtl;
	DevCtl::ScsiInqData			*inqData;
	DevCtl::SCSICAPACITYDATA	*capa;
	INT64						sec_off;
	INT64						sec_num;
	unsigned					iobuf_size;
	BYTE *						iobuf;
	BYTE						*buffer;

public:
	Ext2Lib();
	~Ext2Lib();

	int					OpenExt2(DevCtl & devCtl, DevCtl::ScsiInqData & inq, DevCtl::SCSICAPACITYDATA & capa, INT64 sec_off, INT64 sec_num);
	int					CheckSuperBlock(Ext2Lib::ext2_super_block & superblock);
	TCHAR				*GetOsName(unsigned os, TCHAR * buffer);

private:
	int					ReadBlock(INT64 blkAddr, unsigned blkCount, void * buffer);
	int					Read(INT64 addr, unsigned size, void * buffer);
	virtual int			ReadSector(INT64 adr, unsigned count, unsigned size, void * buffer);
public:
	virtual int			GetBlkSize(void);
	virtual INT64		GetBlkNum(void);
};

#endif
