//////////////////////////////////////////////////////////////////////////
//	"devctl.h"
//------------------------------------------------------------------------
//	$Id: devctl.h 144 2005-07-22 11:16:27Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	DEVCTL_H_INCLUDED
#define	DEVCTL_H_INCLUDED

#include	<windows.h>
#include	<winioctl.h>
#include	<ntddcdrm.h>
#include	<ntddscsi.h>

//#define	DEVCTL_USE_FILEIO

class	DevCtl
{
public:
	typedef __int64 INT64;

	enum rwmode_enum
	{
		RW_READ        = 0x01,
		RW_WRITE       = 0x02,
		RW_SYNC        = 0x04,
		RW_NONBLOCK    = 0x08,
		RW_SHARE_READ  = 0x10,
		RW_SHARE_WRITE = 0x20,
		RW_ASYNC       = 0x40,
		RW_NOBUFFER    = 0x80,
	};
	enum handle_type_enum
	{
		HT_UNKNOWN,
		HT_DISK ,
		HT_DRIVE,
		HT_SCSI ,
	};
	enum filepos_method_enum
	{
		FPOS_BEGIN   = FILE_BEGIN,
		FPOS_END     = FILE_END,
		FPOS_CURRENT = FILE_CURRENT,
	};
#if defined(DEVCTL_USE_FILEIO)
	enum devtype_enum
	{
		DEV_UNKNOWN  ,
		DEV_FLOPPY   ,
		DEV_FIXED    ,
		DEV_REMOVABLE,
		DEV_CDROM    ,
	};
	enum diskformat_enum
	{
		DF_UNKNOWN,
		DF_IBM_FD,
		DF_IBM_HD,
		DF_FMR_HD,
	};
	struct CtlWork
	{
		int				devtype;
		DISK_GEOMETRY	geo;
		INT64			maxBlk;
		unsigned		sectorSize;
	};
#endif

	//////////////////////////////////////////////////////////////////////
	// SCSI
	class ScsiData;
	class ScsiBusData;
	class ScsiInqData;
	struct SCSIINQUIRY;

	class ScsiDevData
	{
	friend class	DevCtl;
	private:
		unsigned		numBus;
		ScsiBusData	   *busData;
	public:
		ScsiDevData();
		~ScsiDevData();
		unsigned		GetBusNum(void) const { return this->numBus; }
		ScsiBusData	*	GetBusData(unsigned n) const { return this->busData == NULL || n >= this->numBus ? NULL : &this->busData[n]; }
	};
	class ScsiBusData
	{
	friend class	DevCtl;
	private:
		unsigned		numUnits;
		unsigned		busId;
		ScsiInqData	   *inqData;
	public:
		ScsiBusData();
		~ScsiBusData();
		unsigned		GetInqNum(void) const { return this->numUnits; }
		unsigned		GetBusId(void) const { return this->busId; }
		ScsiInqData	*	GetInqData(unsigned n) const { return this->inqData == NULL || n >= this->numUnits ? NULL : &this->inqData[n]; }
	};
	class ScsiInqData
	{
	friend class	DevCtl;
	private:
		unsigned		pathId;
		unsigned		targetId;
		unsigned		lun;
		BOOL			devClaimed;
		unsigned		size;
		BYTE *			data;	// SCSIINQUIRY *
	public:
		BOOL			unknownId;
	public:
		ScsiInqData();
		~ScsiInqData();
		unsigned		GetPathId(void) const { return this->pathId; }
		unsigned		GetTargetId(void) const { return this->targetId; }
		unsigned		GetLun(void) const { return this->lun; }
		BOOL			GetClaimed(void) const { return this->devClaimed; }
		unsigned		GetDataSize(void) const { return this->size; }
		BYTE *			GetData(void) const { return this->data; }
		SCSIINQUIRY *	GetInquiryData(void) const { return (SCSIINQUIRY *)this->data; }
		int				Copy(ScsiInqData & src);
	};

	enum scsi_enum
	{
		SCSI_SENSE_LEN = 14,	// Default sense buffer length
		SCSI_CDB6_LEN  = 6,
		SCSI_CDB10_LEN = 10,
		SCSI_CDB12_LEN = 12,
		SCSI_TIMEOUT   = 5000, // 10,
	};
	enum scsi_status_enum
	{
		SCSI_STATUS_GOOD                       = 0x00,
		SCSI_STATUS_CHECK_CONDITION            = 0x02,
		SCSI_STATUS_CONDITION_MET              = 0x04,
		SCSI_STATUS_BUSY                       = 0x08,
		SCSI_STATUS_INTERMEDIATE               = 0x10,
		SCSI_STATUS_INTERMEDIATE_CONDITION_MET = 0x14,
		SCSI_STATUS_RESERVATION_CONFLICT       = 0x18,
		SCSI_STATUS_COMMAND_TERMINATED         = 0x22,
		SCSI_STATUS_QUEUE_FULL                 = 0x28,
	};

#if	defined(_MSC_VER)
#	pragma pack(push,1)
#endif
	struct SPT_BUFFER
	{
		SCSI_PASS_THROUGH	spt;
		ULONG				filler;      // realign buffers to double word boundary
		BYTE				senseArea[SCSI_SENSE_LEN+2];
		UCHAR				buffer[256];
	};
	struct SPT_DIRECT_BUFFER
	{
		SCSI_PASS_THROUGH_DIRECT	spt;
		ULONG						filler;      // realign buffers to double word boundary
		BYTE						senseArea[SCSI_SENSE_LEN+2];
	};
#if	defined(_MSC_VER)
#	pragma pack(pop)
#endif

#	include	"scsihead.h"

	static unsigned	ScsiUint32(const BYTE p[4]) { return (p[0]<<24)+(p[1]<<16)+(p[2]<<8)+p[3]; }
	static unsigned	ScsiUint24(const BYTE p[3]) { return (p[0]<<16)+(p[1]<<8)+p[2]; }
	static unsigned	ScsiUint16(const BYTE p[2]) { return (p[0]<<8)+p[1]; }

private:
	TCHAR		   *devname;
	HANDLE			handle;
	unsigned		htype;		// handle type
	unsigned		rwmode;
#if defined(DEVCTL_USE_FILEIO)
	CtlWork			*ctl;
#endif
	unsigned		scsi_timeout;
	HANDLE			hEvent;
	OVERLAPPED		ol;

public:
	DevCtl();
	~DevCtl();

	int			OpenDevice(const TCHAR * name, unsigned rwmode = RW_READ|RW_WRITE|RW_SHARE_READ);
	int			CloseDevice(void);

	HANDLE			GetHandle(void) const { return this->handle; }
	unsigned		GetHandleType(void) const { return this->htype; }
	int				GetFileType(void);
	int				GetFileSize(LARGE_INTEGER & size);
	INT64			GetFileSize(void);
	int				SetFilePointer(LARGE_INTEGER & pos, int method);
	int				SetFilePointer(INT64 & pos, int method);

	int				DeviceIoControl(DWORD ioctl, void * inbuffer, unsigned insize, void * outbuffer, unsigned outsize, DWORD * byteRet = NULL);
	int				DeviceIoControl(DWORD ioctl, void * inbuffer, unsigned insize, void * outbuffer, unsigned outsize, DWORD * byteRet, OVERLAPPED & overlapped);
	int				DiskIsWritable(void);
	int				DiskGetDriveGeometry(DISK_GEOMETRY & geo);
	int				DiskGetPartitionInfo(PARTITION_INFORMATION & partInfo);
	int				StorageGetMediaTypes(DISK_GEOMETRY & geo);
	int				CdromGetDriveGeometry(DISK_GEOMETRY & geo);
#if (_WIN32_WINNT >= 0x500)
	int				DiskGetLengthInfo(GET_LENGTH_INFORMATION & lenghInfo);
	int				DiskGetDriveGeometryEx(DISK_GEOMETRY_EX & geo);
#endif
	int				StorageGetHotPlugInfo(STORAGE_HOTPLUG_INFO & hi);

	static const TCHAR *	GetMediaTypeName(int mediatype);

#if defined(DEVCTL_USE_FILEIO)
	int				CreateCtlWork(void);
	int				ReadBlk(INT64 pos, unsigned count, void * buffer);
	int				WriteBlk(INT64 pos, unsigned count, const void * buffer);
	int				GetSectorSize(void) const { return this->handle == NULL || this->ctl == NULL ? -1 : (int)this->ctl->sectorSize; }
	int				GetMaxBlk(void) const { return this->handle == NULL || this->ctl == NULL ? -1 : (int)this->ctl->maxBlk; }
	int				GetDiskFormat(void);
#endif

	// SCSI
	static INT64 ScsiGetLBA(BYTE b[4]) { return ((INT64)b[0]<<24) + ((INT64)b[1]<<16) + ((INT64)b[2]<<8) + ((INT64)b[3]); }
	static void  ScsiSetLBA(BYTE b[4], INT64 size) { b[0] = (BYTE)(size>>24); b[1] = (BYTE)(size>>16); b[2] = (BYTE)(size>>8); b[3] = (BYTE)size; }
	int				ScsiGetAddress(SCSI_ADDRESS & scsiAddr);
	int				ScsiGetInquiryData(unsigned size, void * buffer);
	ScsiDevData *	ScsiGetInquiryData(void);
	int				ScsiCreateInqData(DevCtl::ScsiInqData & inqData, int pathId, int targetId, int lun);
	int				ScsiGetCapabilities(IO_SCSI_CAPABILITIES & capa);
	int				ScsiPassThrough(unsigned insize, SCSI_PASS_THROUGH * inspt, unsigned outsize, SCSI_PASS_THROUGH * outspt);
	int				ScsiPassThroughDirect(unsigned insize, SCSI_PASS_THROUGH_DIRECT * inspt, unsigned outsize, SCSI_PASS_THROUGH_DIRECT * outspt);
	int				ScsiPassThroughDirect(unsigned insize, SCSI_PASS_THROUGH_DIRECT * inspt, unsigned outsize, SCSI_PASS_THROUGH_DIRECT * outspt, OVERLAPPED & overlapped);
	int				ScsiCheckCondition(BYTE * senseData);
	int				ScsiInquiry(int pathId, int targetId, int lun, int pageCode, unsigned size, void * buffer);
	int				ScsiTestUnitReady(int pathId, int targetId, int lun);
	int				ScsiLogSense(int pathId, int targetId, int lun, BOOL sp, BOOL ppc, unsigned pc, int pageCode, unsigned point, unsigned size, void * buffer);
	int				ScsiModeSense(int pathId, int targetId, int lun, BOOL blockDesc, unsigned pc, int pageCode, unsigned size, void * buffer);
	int				ScsiModeSense(int pathId, int targetId, int lun, int pageCode, unsigned size, void * buffer) {
					return this->ScsiModeSense(pathId, targetId, lun, FALSE, 0, pageCode, size, buffer); }
	int				ScsiModeSelect(int pathId, int targetId, int lun, BOOL pageFormat, BOOL savePages, unsigned size, const void * buffer);
	int				ScsiReadCapacity(int pathId, int targetId, int lun, unsigned size, void * buffer);
	int				ScsiRead(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, void * buffer);
	int				ScsiRead(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, void * buffer, SPT_DIRECT_BUFFER & spt, OVERLAPPED & overlapped);
	int				ScsiWrite(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, const void * buffer);
	int				ScsiWrite(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, const void * buffer, SPT_DIRECT_BUFFER spt, OVERLAPPED & overlapped);
	int				ScsiVerify(int pathId, int targetId, int lun, INT64 addr, unsigned count);
	int				ScsiVerify(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, const void * buffer, BOOL check);
	int				ScsiSeek(int pathId, int targetId, int lun, INT64 addr);
	int				ScsiReadToc(int pathId, int targetId, int lun, BOOL msf, unsigned size, void * buffer);
	int				ScsiReadToc(int pathId, int targetId, int lun, BOOL msf, DevCtl::SCSITOCDATA & toc);
	int				ScsiReadDefectData(int pathId, int targetId, int lun, unsigned list, unsigned fmt, unsigned size, void * buffer);
	int				ScsiGetConfiguration(int pathId, int targetId, int lun, unsigned rt, unsigned start, unsigned size, void * buffer);
	int				ScsiReadDvdStructure(int pathId, int targetId, int lun, unsigned addr, unsigned format, unsigned layer, unsigned size, void * buffer);
	int				ScsiSetCdSpeed(int pathId, int targetId, int lun, unsigned rc, unsigned speed);

	static TCHAR *	ScsiGetDeviceType(unsigned devType, TCHAR * buffer = NULL);
};


#endif
